# -*- coding: utf-8 -*-
# Generated by Django 1.10.1 on 2017-07-25 09:02
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='KfitsSession',
            fields=[
                ('sid', models.CharField(max_length=32, primary_key=True, serialize=False)),
                ('timestamp', models.DateTimeField(auto_now=True)),
                ('saved_output', models.TextField(null=True)),
            ],
        ),
    ]



import os
import scipy
import time
import math
import json
# this package
import tfitter
from .models import KfitsSession


##########
# CONSTS #
##########
# dimensions of visualisation
FIG_W = 870
FIG_H = 580
SMALL_FIG_W = 520
SMALL_FIG_H = 370
# input file params
MAX_FILE_SIZE = 10*1024*1024
# noise reduction params
MAX_THRESHOLD_SEARCH_ITERATIONS = 10
MAX_NOISE_REMOVAL = 0.4
NOISE_MAX_CONCENTRATION = 1.0


def check_input_file(fnames):
    TO_JSON = True
    fname = fnames[0]
    if not os.access(fname, os.R_OK) or os.path.isdir(fname):
        return TO_JSON, False, "Permission Denied or Directory"
    if os.stat(fname).st_size > MAX_FILE_SIZE:
        return TO_JSON, False, "I am not allowed to read files bigger than %.1f MB" % (MAX_FILE_SIZE / 1048576.)
    data = file(fname,'rb').read()
    if 'XYDATA' not in data:
        return TO_JSON, False, "This file does not seem to contain aggregation data in CSV or TXT format."
    # if we got all the way here...
    return TO_JSON, True, "OK"

def _calibrate_threshold_points(points, w, h):
    w, h = float(w), float(h)
    res = []
    for x,y in points:
        res.append((x/w, 1.-y/h))
    return res

def plot_data(fnames, threshold_points=None, rev_threshold_points=None):
    # prepare return type
    TO_JSON = False
    RETTYPE = 'image/svg+xml'
    # parse parameters
    fname = fnames[0]
    threshold_points = _calibrate_threshold_points(eval(threshold_points[0]), FIG_W, FIG_H) if threshold_points else None
    rev_threshold_points = _calibrate_threshold_points(eval(rev_threshold_points[0]), FIG_W, FIG_H) if rev_threshold_points else None
    # run
    data = scipy.array(tfitter.parse_fluorometer_csv(fname, threshold_points, rev_threshold_points)).T
    return TO_JSON, tfitter.plot_to_svg(data[0], data[1], FIG_W, FIG_H), RETTYPE

def _fit_data(fnames, model, threshold_points=None, rev_threshold_points=None, approx_start=None, search_for_end=True):
    # parse parameters
    fname = fnames[0]
    threshold_points = _calibrate_threshold_points(eval(threshold_points[0]), FIG_W, FIG_H) if threshold_points else None
    rev_threshold_points = _calibrate_threshold_points(eval(rev_threshold_points[0]), FIG_W, FIG_H) if rev_threshold_points else None
    approx_start = float(approx_start[0])/FIG_W if approx_start else 0
    # run
    data = tfitter.parse_fluorometer_csv(fname, threshold_points, rev_threshold_points)
    sim_data, params = tfitter.fit_data(data, approx_start=approx_start*data[-1][0], search_for_end=search_for_end, model=model)
    return data, sim_data, params

def fit_data(fnames, model, threshold_points=None, rev_threshold_points=None, approx_start=None, search_for_end=None):
    # prepare return type
    TO_JSON = False
    RETTYPE = 'image/svg+xml'
    # run
    data, sim_data, params = _fit_data(fnames, model[0], threshold_points, rev_threshold_points, approx_start, search_for_end is not None)
    adata = scipy.array(data).T
    return TO_JSON, tfitter.plot_two_to_svg(adata[0], adata[1], [0.5*i for i in xrange(len(sim_data))], sim_data, FIG_W, FIG_H), RETTYPE

def _clean_data_short(data, sim_data, noise_threshold, noise_only_above=True):
    if noise_only_above:
        fltr = lambda (t,v): v-sim_data[int(t*2)] < noise_threshold
    else:
        fltr = lambda (t,v): abs(v-sim_data[int(t*2)]) < noise_threshold
    new_data = filter(fltr, data)
    adata = scipy.array(new_data).T
    return adata, new_data

def _clean_data(fnames, model, noise_threshold, threshold_points=None, rev_threshold_points=None, approx_start=None, noise_only_above=True, search_for_end=None):
    # parse parameters
    if isinstance(noise_threshold[0], int):
        noise_threshold = noise_threshold[0]
    elif noise_threshold[0].isdigit():
        noise_threshold = int(noise_threshold[0])
    elif noise_threshold[0].replace('.','').isdigit():
        noise_threshold = int(float(noise_threshold[0]))
    else:
        raise ValueError("Noise threshold has illegal value %r" % noise_threshold[0])
    # run fitting
    data, sim_data, params = _fit_data(fnames, model[0], threshold_points, rev_threshold_points, approx_start, search_for_end is not None)
    # find basal level
    baseline = tfitter.find_absolute_baseline(data)
    # clean data
    adata, new_data = _clean_data_short(data, sim_data, noise_threshold, noise_only_above)
    return adata, new_data, baseline, params

def clean_data(fnames, model, noise_threshold, threshold_points=None, rev_threshold_points=None, approx_start=None, noise_only_above=None, search_for_end=None):
    # prepare return type
    TO_JSON = True
    RETM = "OK"
    # run clean
    noise_only_above = noise_only_above is not None
    adata, new_data, baseline, params = _clean_data(fnames, model, noise_threshold, threshold_points, rev_threshold_points, approx_start, noise_only_above, search_for_end)
    # return clean figure and fitting parameters
    return TO_JSON, [tfitter.plot_to_svg(adata[0], adata[1], SMALL_FIG_W, SMALL_FIG_H, classes=['kfigure']), noise_threshold, params.pop('model'), params.pop('t1'), params.pop('t2'), params], RETM

def clean_data_optimise_noise_threshold(fnames, model, threshold_points=None, rev_threshold_points=None, approx_start=None, noise_only_above=None, search_for_end=None):
    # prepare return type
    TO_JSON = True
    RETM = "OK"
    # parse params
    noise_only_above = noise_only_above is not None
    # optimisation loop
    data, sim_data, params = _fit_data(fnames, model[0], threshold_points, rev_threshold_points, approx_start, search_for_end is not None)
    span = max(data, key=lambda (t,v):v)[1] - min(data, key=lambda (t,v):v)[1]
    noise_threshold = span / 3.
    ##last_reduction = 0
    for i in xrange(MAX_THRESHOLD_SEARCH_ITERATIONS):
        adata1, new_data1 = _clean_data_short(data, sim_data, noise_threshold, noise_only_above)
        adata2, new_data2 = _clean_data_short(data, sim_data, noise_threshold/2., noise_only_above)
        diff = set(new_data1) - set(new_data2)
        th2rem = set(data) - set(new_data2)
        # TODO: stop iteration if gain diminishes
        # test if the half threshold is too small
        if len(new_data2) < (1 - MAX_NOISE_REMOVAL) * len(data) or \
                tfitter.get_point_max_concentration(sorted(th2rem), 10)[2] > NOISE_MAX_CONCENTRATION:
            noise_threshold = noise_threshold * 1.5
        # test if halving the threshold is contributing or detrimental
        if len(diff) > 0.7 * MAX_NOISE_REMOVAL * len(data) or \
                tfitter.get_point_max_concentration(sorted(diff), 10)[2] > 0.7 * NOISE_MAX_CONCENTRATION:
            # probably detrimental - go half way
            noise_threshold *= 0.75
        else:
            noise_threshold *= 0.5
        if noise_threshold < 1.:
            # this has gone far enough...
            break
    # final run!
    noise_threshold = int(noise_threshold+0.5)
    adata, new_data, baseline, params = _clean_data(fnames, model, [noise_threshold], threshold_points, rev_threshold_points, approx_start, noise_only_above, search_for_end)
    # return clean figure and fitting parameters
    return TO_JSON, [tfitter.plot_to_svg(adata[0], adata[1], SMALL_FIG_W, SMALL_FIG_H, classes=['kfigure']), noise_threshold, params.pop('model'), params.pop('t1'), params.pop('t2'), params], RETM

def get_clean_data(fnames, model, noise_threshold, output_fnames, threshold_points=None, rev_threshold_points=None, approx_start=None, noise_only_above=None, search_for_end=None):
    # prepare return type
    TO_JSON = False
    RETTYPE = 'text/plain'
    # parse params
    noise_only_above = noise_only_above is not None
    # run clean
    adata, new_data, baseline, params = _clean_data(fnames, model, noise_threshold, threshold_points, rev_threshold_points, approx_start, noise_only_above, search_for_end)
    # prepare file
    new = ['XYDATA,values']
    for x,y in new_data:
        new.append('%f,%f' % (x,y-baseline))
    data = '\n'.join(new)
    # save output to database
    sid = fnames[0].rsplit('.',1)[0][-32:]
    try:
        session = KfitsSession.objects.get(sid=sid)
        session.saved_output = data
        session.kinetic_params = json.dumps(params)
        session.save()
    except Exception:
        pass
    # return clean figure and fitting parameters
    return TO_JSON, data, RETTYPE, output_fnames[0]


#########################
# Non-algorithmic Stuff #
#########################

def get_email_oded():
    return False, '.'.join(('oded','rimon123mail','huji','ac','il')), 'text/plain'

def get_email_dana():
    return False, '.'.join(('danare123mail','huji','ac','il')), 'text/plain'



import os, sys
try:
    from kfits.afitter import *
except ImportError:
    # in case the package is not properly installed, try to work locally
    import inspect
    currentdir = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
    parentdir = os.path.dirname(currentdir)
    sys.path.insert(0,parentdir)
    from afitter import *

#########
# UTILS #
#########

SVG_TEMPLATE = """<?xml version="1.0" encoding="UTF-8" standalone="yes"?>
<svg xmlns="http://www.w3.org/2000/svg" width="%(width)d" height="%(height)d" %(classes)s viewBox="0 0 %(width)d %(height)d" preserveAspectRatio="none">
    <defs>
        <style type="text/css" />
    </defs>
    <g>
        %(gdata)s
    </g>
</svg>"""
SVG_PATH_TEMPLATE = """<path style="fill:none; stroke:%(color)s; stroke-width:1px"
    d="%(path)s" />"""

def plot_to_svg(xx, yy, w, h, color="#000000", classes=[]):
    min_x, max_x = min(xx), max(xx)
    min_y, max_y = min(yy), max(yy)
    norm_xx = (scipy.array(xx) - min_x) / (max_x - min_x) * w
    norm_yy = h - ((scipy.array(yy) - min_y) / (max_y - min_y) * h)
    retval = ["M %f %f" % (norm_xx[0], norm_yy[0])]
    for i in xrange(1, len(norm_xx)):
        retval.append('L %f %f' % (norm_xx[i], norm_yy[i]))
    path = SVG_PATH_TEMPLATE % dict(color = color,
                                    path = ' '.join(retval))
    return SVG_TEMPLATE % dict(width = w,
                               height = h,
                               gdata = path,
                               classes = ('class="%s"' % ' '.join(classes)) if classes else '')

def plot_two_to_svg(xx1, yy1, xx2, yy2, w, h, color1="#000000", color2="#ff4414"):
    min_x, max_x = min(min(xx1),min(xx2)), max(max(xx1),max(xx2))
    min_y, max_y = min(min(yy1),min(yy2)), max(max(yy1),max(yy2))
    norm_xx1 = (scipy.array(xx1) - min_x) / (max_x - min_x) * w
    norm_yy1 = h - ((scipy.array(yy1) - min_y) / (max_y - min_y) * h)
    norm_xx2 = (scipy.array(xx2) - min_x) / (max_x - min_x) * w
    norm_yy2 = h - ((scipy.array(yy2) - min_y) / (max_y - min_y) * h)
    retval = ["M %f %f" % (norm_xx1[0], norm_yy1[0])]
    for i in xrange(1, len(norm_xx1)):
        retval.append('L %f %f' % (norm_xx1[i], norm_yy1[i]))
    path1 = SVG_PATH_TEMPLATE % dict(color = color1,
                                     path = ' '.join(retval))
    retval = ["M %f %f" % (norm_xx2[0], norm_yy2[0])]
    for i in xrange(1, len(norm_xx2)):
        retval.append('L %f %f' % (norm_xx2[i], norm_yy2[i]))
    path2 = SVG_PATH_TEMPLATE % dict(color = color2,
                                     path = ' '.join(retval))
    return SVG_TEMPLATE % dict(width = w,
                               height = h,
                               gdata = path1 + '\n' + path2,
                               classes = '')



#!/usr/bin/env python
import os
import sys

def main(args, settings="afgui.settings"):
    os.environ.setdefault("DJANGO_SETTINGS_MODULE", settings)
    try:
        from django.core.management import execute_from_command_line
    except ImportError:
        # The above import may fail for some other reason. Ensure that the
        # issue is really that Django is missing to avoid masking other
        # exceptions on Python 2.
        try:
            import django
        except ImportError:
            raise ImportError(
                "Couldn't import Django. Are you sure it's installed and "
                "available on your PYTHONPATH environment variable? Did you "
                "forget to activate a virtual environment?"
            )
        raise
    execute_from_command_line(args)

def run_with_default_settings():
    # find the directory of `manage.py`
    import inspect
    currentdir = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
    os.chdir(currentdir)
    sys.path.insert(0, currentdir)
    # run django server (noreload is necessary, because otherwise this command fails when not running from the same path as `manage.py`
    main(['', 'runserver', '--noreload'], "kfits.afgui.afgui.settings")
##    # ALTERNATIVE WAY
##    import subprocess
##    import webbrowser
##    subprocess.Popen(['python', 'manage.py', 'runserver'], cwd=currentdir)
##    webbrowser.open_new_tab('http://127.0.0.1:8000/fitter')

def run_as_server():
    # find the directory of `manage.py`
    import inspect
    currentdir = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
    os.chdir(currentdir)
    sys.path.insert(0, currentdir)
    # get ip and port from user
    ip = raw_input("Enter IP to serve on (use 0.0.0.0 for all interfaces, 127.0.0.1 for local loop interface only): ")
    port = raw_input("Enter port to serve on (use 8000 if you don't have another preference): ")
    if not port.isdigit or int(port) < 1 or int(port) > 65535:
        print >> sys.stderr, "Port must be a number between 1-65535."
        port = raw_input("Try port again: ")
    addr = '%s:%s' % (ip, port)
    # set environment variable for new address
    if ip == '0.0.0.0':
        client_addr = '127.0.0.1:%s' % port
    else:
        client_addr = addr
    os.environ.setdefault("DJANGO_SERVER_ADDRESS", client_addr)
    # run django server (noreload is necessary, because otherwise this command fails when not running from the same path as `manage.py`
    main(['', 'runserver', '--noreload', addr], "kfits.afgui.afgui.settings")


if __name__ == "__main__":
    main(sys.argv)



#!/usr/bin/env python
import os
import sys
import math
import scipy
from scipy.optimize import curve_fit
# for debugging
import inspect
# this package
import kmodels

#########
# UTILS #
#########

def _debug(d):
    last_func_name = _debug.__dict__.setdefault('last_func', None)
    func_name = inspect.getouterframes(inspect.currentframe(), 2)[1][3]
    if last_func_name != func_name:
        print >> sys.stderr
        print >> sys.stderr, func_name
    _debug.__dict__['last_func'] = func_name
    for k,v in d.iteritems():
        print >> sys.stderr, '\t',k,'=',repr(v)

def check_threshold(point, threshold_points):
    """
    @param point: Point to check
    @type point: 2-tuple
    @param threshold_points: Points on the graph defining thresholds for noise (optional, default: none)
    @type threshold_points: list of 2-tuples

    @return: Is the point above the line defined by the threshold points? (if it is neither above nor below - but left or right to the line - will return None)
    @rtype: bool or None
    """
    if point[0] <= threshold_points[0][0] or point[0] > threshold_points[-1][0]:
        return None
    for i in xrange(len(threshold_points)-1):
        x1, y1 = map(float, threshold_points[i])
        x2, y2 = map(float, threshold_points[i+1])
        if point[0] > x1 and point[0] <= x2:
            rel = (point[0] - x1) / (x2 - x1)
            return point[1] <= y1 + (y2 - y1) * rel
    raise RuntimeError("How did we get here?!")

def get_point_max_concentration(data, min_window=3):
    """
    @param data: Filtered aggregation data
    @type data: List of 2-tuples
    @param min_window: Minimum number of points on which we can calculate concentrations [default: 3]
    @type min_window: int

    @return: The maximum concentration of points found (at points/sec), and where they were found - (start, end, concentration)
    @rtype: tuple (int, int, float)
    """
    # check params
    if len(data) <= min_window:
        return (0, len(data), 0., 0.)
    # init
    top = (0, 0, 0., 0.)
    start = 0
    end = min_window
    cur_con = float(end) / (data[end-1][0] - data[start][0])
    cur_score = float(end)**0.5 * cur_con
    while end < len(data):
        # calculate scores for adding one point
        advcon = float(end-start+1) / (data[end][0] - data[start][0])
        # adjust score to number of points
        adv = float(end-start+1)**0.5 * advcon
        # should we advance the end index?
        if adv >= cur_score:
            # yes!
            cur_score = adv
            cur_con = advcon
            end += 1
        else:
            # no - but have we reached a top scoring set?
            if cur_score > top[3]:
                # new record!
                top = (start, end, cur_con, cur_score)
            end += 1
            start = end - min_window
    # have we reached a top scoring set at the end?
    if cur_score > top[3]:
        # new record!
        return (start, end, cur_con)
    else:
        return top[:3]


################
# FILE PARSERS #
################

def parse_fluorometer_csv(path, threshold_points=None, rev_threshold_points=None, debug=False):
    """
    Parse a CSV or TXT file created by the fluorometer.

    @param path: The path of the file
    @type path: str
    @param threshold_points: Points on the graph defining lower thresholds for noise (optional, default: none)
    @type threshold_points: list of 2-tuples
    @param rev_threshold_points: Points on the graph defining upper thresholds for noise (optional, default: none)
    @type rev_threshold_points: list of 2-tuples

    @return: Aggregation Data
    @rtype: List of 2-tuples
    """
    if not os.path.isfile(path):
        raise TypeError("%s is not a path to a regular file" % path)
    elif not os.access(path, os.R_OK):
        raise TypeError("Do not have reading permissions for %s" % path)
    else:
        data = file(path, 'r').read()
    if debug:
        _debug(dict(path=path, threshold_points=threshold_points, rev_threshold_points=rev_threshold_points))
    retval = []
    started = False
    ended = False
    ncomma, ntab, nlines = data.count(','), data.count('\t'), data.count('\n')
    if ncomma < ntab and ntab > nlines / 2.:
        sep = '\t'
    else:
        sep = ','
    for line in data.splitlines():
        if 'XYDATA' in line:
            started = True
        elif started and not ended:
            if line.strip() == "":
                ended = True
            else:
                retval.append(tuple(map(float, line.split(sep))))
    if not threshold_points and not rev_threshold_points:
        return retval
    else:
        tmp = scipy.array(retval)
        min_x, min_y = tmp.min(0)
        max_x, max_y = tmp.max(0)
        x_span = max_x - min_x
        y_span = max_y - min_y
        if debug:
            _debug({"points before filtering\n": len(retval)})
        if threshold_points:
            new_tps = []
            for x,y in threshold_points:
                new_tps.append((min_x+x*x_span, min_y+y*y_span))
            retval = filter(lambda point: check_threshold(point, new_tps) in (True, None), retval)
            if debug:
                _debug({"points after filtering by green line\n": len(retval)})
        if rev_threshold_points:
            new_tps = []
            for x,y in rev_threshold_points:
                new_tps.append((min_x+x*x_span, min_y+y*y_span))
            retval = filter(lambda point: check_threshold(point, new_tps) in (False, None), retval)
            if debug:
                _debug({"points after filtering by red line\n": len(retval)})
        return retval


################
# DATA FITTING #
################

def closest_t(data, t):
    """
    Return the index of the line with the closest t (time) value to the given t.
    Assumes the data is sorted by time. Returns None in case of failure.
    """
    if data is None or len(data) == 0:
        return
    cur = int(t*2)
    if cur >= len(data) - 1:
        cur = len(data) - 2
    if cur < 0:
        cur = 0
    inc = (-1,1)[data[cur][0] < t]
    for i in xrange(len(data)):
        if abs(data[cur][0] - t) < abs(data[cur-1][0] - t) and \
           abs(data[cur][0] - t) < abs(data[cur+1][0] - t):
            return cur
        else:
            cur += inc
            if cur == 0:
                return
            elif cur == len(data)-1:
                return cur
    return None

def find_absolute_baseline(data, debug=False):
    """
    Find the absolute baseline of the measurement.

    @param data: The aggregation data
    @type data: List of 2-tuples
    @param debug: Show debug information? (default: no)
    @type debug: bool

    @return: Basal value
    @rtype: float
    """
    # find first 30 seconds of measurement
    tmp_data = scipy.array(data[:60])
    mean = tmp_data.mean(0)[1]
    stdev = tmp_data.std(0)[1]
    # filter out "bad" data points
    filtered_data = scipy.array(filter(lambda (t,v): abs(v-mean) < stdev, tmp_data))
    better_mean = filtered_data.mean(0)[1]
    if debug:
        _debug(dict(mean=mean, stdev=stdev, better_mean=better_mean, better_stdev=filtered_data.std(0)[1]))
    return better_mean

def find_aggregation_initiation(data, approx=120, debug=False):
    """
    Find the beginning of the aggregation process, i.e. the point in time
    where the actual experiment began, i.e. the time when the substrate was
    added to the mix.

    @param data: The aggregation data
    @type data: List of 2-tuples
    @param approx: Approximate time to search around (default: 120 sec)
    @type approx: number
    @param debug: Show debug information? (default: no)
    @type debug: bool

    @return: Basal value and Time of aggregation beginning (baseline, t1)
    @rtype: tuple (float, float)
    """
    if debug:
        _debug(dict(data_len=len(data), beginning=data[:10], approx=approx))
    RANGE_WIDTH = 40
    # find last 30 seconds before approximate init
    end_idx = closest_t(data, approx)
    if end_idx is None:
        end_idx = len(data) - 1
    if debug:
        _debug({'end index':end_idx})
    tmp_data = scipy.array(data[end_idx - 60 if end_idx > 60 else 0:end_idx + 1])
    mean = tmp_data.mean(0)[1]
    stdev = tmp_data.std(0)[1]
    # filter out "bad" data points
    filtered_data = scipy.array(filter(lambda (t,v): abs(v-mean) < stdev, tmp_data))
    better_mean = filtered_data.mean(0)[1]
    better_stdev = filtered_data.std(0)[1]
    if debug:
        _debug(dict(mean=mean, stdev=stdev, better_mean=better_mean, better_stdev=better_stdev))
    # locate t1
    best_idx = end_idx
    range_width = RANGE_WIDTH if RANGE_WIDTH < end_idx else end_idx
    for i in xrange(end_idx - range_width, end_idx + range_width + 1):
        smean = sum([x[1] for x in data[i:i+10]]) / 10.
        if smean > better_mean + 1.5 * better_stdev:
            best_idx = i
            break
    if debug:
        _debug(dict(end_idx=end_idx, best_idx=best_idx, smean=smean))
    return better_mean, data[best_idx][0]

def find_apparent_maximum(data, range_to_fit=120, minimal_t2=350., debug=False):
    """
    Find the point where aggregation reached its apparent maximum.
    This is often long before the end of the trace, as larger aggregates
    cannot be measured in some methods and effectively appear as plateaux
    or even slight decrease in measured values (for example in scattering).

    @param data: The aggregation data
    @type data: List of 2-tuples
    @param range_to_fit: How many timepoints at the end of the trace to fit with line (default: 120, i.e. 2 min)
    @type range_to_fit: int
    @param minimal_t2: Minimal value for t2, important for low-aggregation samples (default: 350s)
    @type minimal_t2: number
    @param debug: Show debug information? (default: no)
    @type debug: bool

    @return: The maximal point (t2, v_t2)
    @rtype: tuple (float, float)
    """
    DEVIATIONS_FROM_LINE = 10. 
    # fit final minute with line
    to_fit = data[-range_to_fit:]
    popt, pcov = curve_fit(lambda x, a, b: a*x+b, [x[0] for x in to_fit], [x[1] for x in to_fit])
    a, b = popt
    if debug:
        _debug(dict(fit_a=a, fit_b=b))
    ##perr = scipy.sqrt(scipy.diag(pcov))
    # calculate standard distance from line
    diffs = scipy.array([abs((a*x[0]+b)-x[1]) for x in data])
    fit_dev = diffs[-range_to_fit:].std()
    # forecast backwards using line, find point of divergence from linearity
    bins = [sum(diffs[i:i+10])/10. for i in xrange(0,len(diffs),10)]
    point_of_divergence = None
    while point_of_divergence is None or data[point_of_divergence][0] < minimal_t2:
        latest_divergent_bin = 0
        for i in xrange(len(bins)-2,-1,-1):
            if bins[i] > DEVIATIONS_FROM_LINE*fit_dev:
                latest_divergent_bin = i
                break
        if debug:
            _debug(dict(latest_divergent_bin=latest_divergent_bin, num_bins=len(bins)))
        for i in xrange(9,-1,-1):
            idx = latest_divergent_bin*10 + i
            if diffs[idx] > DEVIATIONS_FROM_LINE*fit_dev:
                point_of_divergence = idx
        DEVIATIONS_FROM_LINE /= 2.
    if debug:
        _debug(dict(point_of_divergence=point_of_divergence, retval=data[point_of_divergence]))
    return data[point_of_divergence]

def fit_aggregation_kinetics(data, baseline, t1, t2, apparent_max, fit_func, init_values, debug=False):
    """
    Fit the main aggregation period data with (t*vmax)/(t+t1/2)) kinetics.

    @param data: The aggregation data
    @type data: List of 2-tuples
    @param baseline: The calculated baseline of the curve
    @type baseline: float
    @param t1: Time of the beginning of the aggregation kinetics
    @type t1: number
    @param t2: Time of the end of the aggregation kinetics
    @type t2: number
    @param apparent_max: The apparent maximum value
    @type apparent_max: float
    @param fit_func: The fitting function; receives t (time, shifted to start from 0) and any parameters that should be optimised
    @type fit_func: function
    @param init_values: Initial values for all parameters of fit_func except t
    @type init_values: tuple of numbers
    @param debug: Show debug information? (default: no)
    @type debug: bool

    @return: popt and perr
    @rtype: tuple, tuple
    """
    to_fit = data[closest_t(data,t1):closest_t(data,t2)+1]
    popt, pcov = curve_fit(fit_func, [x[0]-t1 for x in to_fit], [x[1]-baseline for x in to_fit], p0=init_values, maxfev=10000)
    if debug:
        _debug(dict(popt=popt, pcov=pcov))
    perr = scipy.sqrt(scipy.diag(pcov))
    ##return [x[0]-t1 for x in to_fit], [x[1]-baseline for x in to_fit]
    return popt, perr

def choose_best_model(data, baseline, approx_start, t1, t2, apparent_max, fit_pairs, debug=False):
    """
    Fit the main aggregation period data with (t*vmax)/(t+t1/2)) kinetics.

    @param data: The aggregation data
    @type data: List of 2-tuples
    @param baseline: The calculated baseline of the curve
    @type baseline: float
    @param approx_start: Approximate time of beginning of kinetics, estimated by the user
    @type approx_start: number
    @param t1: Time of the beginning of the aggregation kinetics
    @type t1: number
    @param t2: Time of the end of the aggregation kinetics
    @type t2: number
    @param apparent_max: The apparent maximum value
    @type apparent_max: float
    @param fit_pairs: Dictionary of names pointing to 2-tuples, each comprised of a fitting function and initial values
    @type fit_pairs: dict (str:tuple)
    @param debug: Show debug information? (default: no)
    @type debug: bool

    @return: The best fit_pair's name
    @rtype: str
    """
    to_fit = data[closest_t(data,t1):closest_t(data,t2)+1]
    x_vals = [x[0]-t1 for x in to_fit]
    y_vals = [x[1]-baseline for x in to_fit]
    # fitting loop
    best = None
    best_score = None
    for name, (fit_func, init_values, param_names, human_name) in fit_pairs.iteritems():
        init_values = init_values(apparent_max, baseline)
        try:
            popt, pcov = curve_fit(fit_func, x_vals, y_vals, p0=init_values, maxfev=10000)
        except Exception, e:
            print >> sys.stderr, "Encountered error trying to fit data with %r: %s" % (name, e)
            continue
        score = sum([(y_vals[i] - fit_func(x_vals[i], *popt))**2 for i in xrange(len(y_vals)) if to_fit[i][0]>approx_start])
        if debug:
            _debug(dict(approx_start=approx_start, model_name=name, score=score, old_score=sum([abs(y_vals[i] - fit_func(x_vals[i], *popt)) for i in xrange(len(y_vals))])))
        if best is None or score < best_score:
            best = name
            best_score = score
    return best


#####################
# FITTING FUNCTIONS #
#####################

FITTING_PAIRS = kmodels.get_models()


#############
# INTERFACE #
#############

def fit_data(data, model='auto', approx_start=120, search_for_end=True, debug=False):
    # find initiation
    baseline, t1 = find_aggregation_initiation(data, approx=approx_start, debug=debug)
    # find plateau
    if search_for_end:
        t2, apparent_max = find_apparent_maximum(data, minimal_t2=(data[-1][0]+approx_start)/2, debug=debug)
    else:
        t2, apparent_max = data[-1]
    # choose model
    if model == 'auto':
        model = choose_best_model(data, baseline, approx_start, t1, t2, apparent_max, FITTING_PAIRS, debug=True)
        if model is None:
            raise RuntimeError("None of the models successfully fit the data!")
    fit_pair = FITTING_PAIRS[model]
    # fit kinetics
    popt, perr = fit_aggregation_kinetics(data, baseline, t1, t2, apparent_max, fit_pair[0], fit_pair[1](apparent_max, baseline), debug=debug)
    if debug:
        d = dict(zip(inspect.getargspec(fit_pair[0])[0][2:], ['%.3f +- %.3f' % (popt[i], perr[i]) for i in xrange(len(popt))]))
        _debug(d)
    # fit rest of data with straight line
    end_of_agg_curve = baseline + fit_pair[0](t2-t1, *popt)
    to_fit = data[closest_t(data,t2):]
    a, pcov = curve_fit(lambda x, a: a*x, [x[0]-t2 for x in to_fit], [x[1]-end_of_agg_curve for x in to_fit])
    a = float(a)

    # return fitted (simulated) data
    simulated_data = [baseline for x in xrange(int(t1*2))] + \
                     [baseline + fit_pair[0](x/2., *popt) for x in xrange(int((t2-t1)*2))] + \
                     [(a*((x/2.)-t2)+end_of_agg_curve) for x in range(int(t2*2.)-1, int(2*max(data)[0]))]
    # prepare return value
    params = dict(model = model, t1 = t1, t2 = t2)
    for i, param_name in enumerate(fit_pair[2]):
        params[param_name] = '%.3f &plusmn; %.3f' % (popt[i], perr[i])
    return simulated_data, params


##########################
# COMMAND-LINE INTERFACE #
##########################

def main(args):
    if len(args) < 1 or len(args) > 2:
        print >> sys.stderr, "Usage: %s <csv file path> (<approx start>)" % sys.argv[0]
        return 1
    try:
        from matplotlib import pyplot
    except ImportError, e:
        print >> sys.stderr, "The command line interface requires installation of the python matplotlib package."
        return 1
    data = parse_fluorometer_csv(args[0], debug=True)
    _debug({'data length':len(data)})
    if len(args) > 1:
        sim_data, params = fit_data(data, approx_start=int(args[1]), debug=True)
    else:
        sim_data, params = fit_data(data, debug=True)
    for item in params.iteritems():
        print '%s: %s' % item
    pyplot.plot([x[0] for x in data], [x[1] for x in data])
    pyplot.plot([0.5*i for i in xrange(len(sim_data))], sim_data)
    pyplot.show()
    return 0

if __name__ == '__main__':
    sys.exit(main(sys.argv[1:]))



#!/usr/bin/env python
import os
import sys
import afitter as fitter


DEFAULT_THRESHOLD = 25


def main(args):
    if len(args) < 1:
        print >> sys.stderr, "Usage: %s <csv file path> (<threshold>)" % sys.argv[0]
        return 1
    threshold = DEFAULT_THRESHOLD if len(args) < 2 else int(args[1])
    # read & fit data
    data = fitter.parse_fluorometer_csv(args[0])
    sim_data = fitter.fit_data(data)
    # find basal level
    baseline = fitter.find_absolute_baseline(data)
    # run fitting
    sim_data, params = fitter.fit_data(data)
    # clean data
    print 'threshold:', threshold
    new_data = filter(lambda (t,v): v-sim_data[int(t*2)] < threshold, data)
    # write to new file
    new_path = args[0].rsplit('.',1)[0] + '_clean.csv'
    w = file(new_path, 'w')
    w.write('Time (s),Light Scattering (A.U.)\n')
    w.writelines(['%.1f,%f\n' % (point[0], point[1]-baseline) for point in new_data])
    w.close()
    print "\nWRITTEN CLEAN DATA TO %s" % new_path
    return 0


if __name__ == '__main__':
    sys.exit(main(sys.argv[1:]))



import os
import mimetypes
import json
import time
import datetime
import random
import hashlib
import json
import inspect
# django
from django.shortcuts import render
from django import http
from django import template
from .forms import UploadFileForm
from .models import KfitsSession
# this module
import fbackend


TMP_FILENAME = 'kfits_%s.txt'


def _get_tmp_dir():
    return os.environ.get('TEMP', os.environ.get('TMP', '/tmp'))

# Create your views here.
def index(request):
    # create index
    tmplt = template.loader.get_template('fitter/index.htm')
    parent_dir = os.path.dirname(os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe()))))
    return http.HttpResponse(tmplt.render(dict(tmp_file = '',
                                               example1 = os.path.join(parent_dir, 'example1.csv'),
                                               example2 = os.path.join(parent_dir, 'example2.csv'),
                                               example3 = os.path.join(parent_dir, 'example3.csv'),
                                               model_choice = [(model, fbackend.tfitter.FITTING_PAIRS[model][3]) for model in fbackend.tfitter.FITTING_PAIRS.keys()],
                                               width = fbackend.FIG_W,
                                               height = fbackend.FIG_H,
                                              ), request))

def test(request):
    tmplt = template.loader.get_template('fitter/test.htm')
    sessions = KfitsSession.objects.all()
    return http.HttpResponse(tmplt.render({'sessions': '<br>'.join(['%s (%s): %d bytes, %s' % (ses.sid, ses.timestamp.ctime(), len(ses.saved_output) if ses.saved_output else 0, ses.kinetic_params) for ses in sessions])}, request))

def bootstrap(request):
    response = file(os.path.join(os.getcwd(), 'fitter/templates', request.path.strip('/')),'rb').read()
    return http.HttpResponse(response, mimetypes.guess_type(request.path)[0])

def backend(request):
    if request.GET.has_key("function") and hasattr(fbackend, request.GET['function']):
        params = dict(request.GET)
        func = getattr(fbackend, params.pop('function')[0])
        # run function
        params.pop('_', None)
        res = func(**params)
        if res[0]:
            new_res = res[1:]
            if len(new_res) == 1:
                new_res = new_res[0]
            # return output
            return http.HttpResponse(json.dumps(new_res), "application/json")
        else:
            # return output
            response = http.HttpResponse(res[1], res[2])
            if len(res) > 3:
                response['Content-Disposition'] = 'attachment; filename="%s"' % res[3];
            return response
    else:
        return http.HttpResponseBadRequest()

def upload_text(request):
    if request.method == 'POST':
        form = UploadFileForm(request.POST, request.FILES)
        if form.is_valid():
            # get client ip
            forwarded_for = request.META.get('HTTP_X_FORWARDED_FOR')
            client_ip = forwarded_for.split(',')[0] if forwarded_for else request.META.get('REMOTE_ADDR')
            # randomize session id
            full_session_details = '%s%.3f%.8f' % (client_ip, time.time(), random.random())
            session_id = hashlib.md5(full_session_details).hexdigest()
            # create session
            session = KfitsSession(sid = session_id)
            session.save()
            # delete old sessions
            old_sessions = KfitsSession.objects.filter(timestamp__lte=datetime.datetime.now()-datetime.timedelta(1,0,0))
            for oldses in old_sessions:
                # remove temporary file
                sfname = os.path.join(_get_tmp_dir(), TMP_FILENAME % oldses.sid)
                try:
                    os.remove(sfname)
                except OSError:
                    pass
                # delete entry from database
                oldses.delete()
            # make form
            fname = os.path.join(_get_tmp_dir(), TMP_FILENAME % session_id)
            f = file(fname, 'w')
            for chunk in request.FILES['fdata'].chunks():
                f.write(chunk)
            return http.HttpResponse(json.dumps(fname), "application/json")
    else:
        return render(request, 'upload.htm', {'form': UploadFileForm()})
    return http.HttpResponse(json.dumps(False), "application/json")

def get_past_results(request):
    if request.GET.has_key("sid"):
        sid = request.GET['sid']
        try:
            session = KfitsSession.objects.get(sid=sid)
        except Exception:
            return http.HttpResponseNotFound('<h1>Cannot find session %s</h1>It was either deleted or never existed.' % sid)
        if request.GET.has_key("download"):
            data = session.saved_output
            response = http.HttpResponse(data, 'text/plain')
            response['Content-Disposition'] = 'attachment; filename="%s.csv"' % sid
            return response
        else:
            if session.kinetic_params:
                tmplt = template.loader.get_template('fitter/get.htm')
                return http.HttpResponse(tmplt.render(dict(kparams = [dict(zip(['name','value'],param)) for param in sorted(json.loads(session.kinetic_params).items())],
                    sid = sid)))
            else:
                return http.HttpResponseNotFound('<h1>Cannot find results for session %s</h1>The session exists, but cleaning results were not saved by the user.' % sid)
    else:
        tmplt = template.loader.get_template('fitter/get_nosid.htm')
        return http.HttpResponse(tmplt.render())



from __future__ import unicode_literals

from django.db import models

# Create your models here.

class KfitsSession(models.Model):
    sid = models.CharField(max_length=32, primary_key=True)
    timestamp = models.DateTimeField(auto_now=True)
    saved_output = models.TextField(null=True)
    kinetic_params = models.TextField(null=True)



from django.conf import urls
from . import views

urlpatterns = [
    urls.url(r'^$', views.index, name='index'),
    urls.url(r'^index.html?$', views.index, name='index'),
    urls.url(r'^test.html?$', views.test, name='test'),
    urls.url(r'^bootstrap/.*$', views.bootstrap, name='bootstrap'),
    urls.url(r'^backend$', views.backend, name='fbackend'),
    urls.url(r'^upload_text.html?$', views.upload_text, name='upload_text'),
    urls.url(r'^get.htm?$', views.get_past_results, name='get_past_results'),
]



import math
import scipy


FIT_BASIC_INIT = lambda apparent_max, baseline: (apparent_max-baseline, 12)
FIT_BASIC_PARAM_NAMES = ('v<sub>max</sub>', 't<sub>&#189;</sub>')
def fit_basic(t, vmax, thalf):
    return (t * vmax) / (t + thalf)

FIT_CAMBRIDGE_INIT = lambda apparent_max, baseline: (apparent_max-baseline, 1.1, 1000)
FIT_CAMBRIDGE_PARAM_NAMES = ('v<sub>max</sub>', 'n<sub>c</sub>', 'k')
class CambridgeFit(object):
    def __init__(self, m_total):
        self.mtot = m_total

    def fit_cambridge(self, t, vmax, nc, k):
        if nc < 0 or nc > 10 or k < 0 or k > 1e12:
            return 0
        ins = (nc * k * self.mtot**nc)**0.5
        try:
            outs = scipy.vectorize(math.cosh)(ins * t)
        except OverflowError, e:
            return 0
        return vmax * (1 - outs ** (-2./nc))


def get_models():
    return dict(basic = (fit_basic, \
                         FIT_BASIC_INIT, \
                         FIT_BASIC_PARAM_NAMES, \
                         'Basic One-Site Binding'),
                nucleation_elongation = (CambridgeFit(75e-6).fit_cambridge, \
                                         FIT_CAMBRIDGE_INIT, \
                                         FIT_CAMBRIDGE_PARAM_NAMES, \
                                         'Nucleation Elongation'),
                )




# -*- coding: utf-8 -*-
# Generated by Django 1.10.1 on 2017-08-02 13:27
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('fitter', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='kfitssession',
            name='kinetic_params',
            field=models.TextField(null=True),
        ),
    ]



"""
WSGI config for afgui project.

It exposes the WSGI callable as a module-level variable named ``application``.

For more information on this file, see
https://docs.djangoproject.com/en/1.10/howto/deployment/wsgi/
"""

import os

from django.core.wsgi import get_wsgi_application

os.environ.setdefault("DJANGO_SETTINGS_MODULE", "afgui.settings")

application = get_wsgi_application()

# get server addr
addr = os.environ.get("DJANGO_SERVER_ADDRESS", '127.0.0.1:8000')
# open browser
try:
    import webbrowser
    webbrowser.open_new_tab('http://%s/fitter' % addr)
except:
    import sys
    print >> sys.stderr, "Warning: Couldn't open web browser client"



import platform
from setuptools import setup

top_fnames = ['LICENSE','README','README.md']
# overcome annoying compatability bug
if platform.system() != 'Windows':
    top_fnames = ['.'+x for x in top_fnames]

setup(name = 'kfits',
      version = '1.1',
      description = 'Kinetics Fitting Software',
      long_description = file('README.md','r').read(),
      author = 'Oded Rimon',
      author_email = '%s%s%s' % ('oded.rimon', chr(32*2), '.'.join(('mail','huji','ac','il'))),
      url = 'https://github.com/odedrim/kfits',
      install_requires = {'django', 'scipy'},
      package_dir = {'kfits': ''},
      packages = ['kfits',
                  'kfits.afgui',
                  'kfits.afgui.afgui',
                  'kfits.afgui.fitter',
                  'kfits.afgui.fitter.migrations'],
      package_data = {'kfits': top_fnames,
                      'kfits.afgui': ['db.sqlite3', 'example*.csv'],
                      'kfits.afgui.afgui': ['LICENSE'],
                      'kfits.afgui.fitter': ['templates/*.*', 'templates/fitter/*.*', 'templates/fitter/bootstrap/*.*', 'templates/fitter/bootstrap/assets/*.*', 'templates/fitter/bootstrap/css/*', 'templates/fitter/bootstrap/fonts/*', 'templates/fitter/bootstrap/js/*', 'templates/fitter/bootstrap/assets/brand/*', 'templates/fitter/bootstrap/assets/css/*.*', 'templates/fitter/bootstrap/assets/flash/*', 'templates/fitter/bootstrap/assets/img/*', 'templates/fitter/bootstrap/assets/js/*.*', 'templates/fitter/bootstrap/assets/css/src/*', 'templates/fitter/bootstrap/assets/js/src/*', 'templates/fitter/bootstrap/assets/js/vendor/*']},
      entry_points = {'console_scripts': ['kfits = kfits.afgui.manage:run_with_default_settings',
                                          'kfits_server = kfits.afgui.manage:run_as_server']}
      )

# list of package_data for kfits.afgui.fitter was partially created using the python code:
# sum([['%s/%s/*' % (x[0].replace(os.path.sep,'/'), y) for y in x[1]] for x in os.walk('templates/fitter/bootstrap')], [])
